############################
#
# missionobjvar.py
#
############################

# See http://www.freedomreborn.net/invision/index.php?showtopic=26722&hl=
# for documentation.  Or search the Scripting forum on Freedom Reborn for
#     +documentation +general +Mission_SetVar 

# Mission and Object Variables:
# Storing an arbitrary variable that persists through game saves and
# reloads as we do with the {Mission,Object}_{Get,Set}Attr() functions,
# but not limited to floats.  I am calling these 'long variables'.
# E.g., we use these like
#   myThugs = ['thug1','thug2','thug3']
#   Mission_SetVar('somethugs',myThugs)
#   ...
#   myThugs = Mission_GetVar('somethugs')
# And so on.
# 2004/09/30
# The current implementation uses a hashed variable name lookup to make
# the process more memory efficient and allow for better error-checking.
# It's probably slower, but I haven't seen a practical difference.
# A side benefit of this approach is the availability of functions like
# Mission_VarExists(varName), Object_VarExists(object,varName),
# Mission_GetAllVarNames(), and Object_GetAllVarNames(object).  Also added file
# reading and writing of the variables for inter-mission persistence.
# 2004/11/18
# Made independent of cshelper.py so that it can be loaded in any order.
# 2004/11/24
# Fixed a typo resulting in an aliasing bug that was potential trouble
# when one object had >=10 vars of >=48 bytes.
# 2005/04/02
# Some speed an memory improvements made.
# 2005/07/02
# Altered variable value-to-string to use repr() instead of str().
# On module load, added check for badly exported long ints, like
# js.SCSTATE_BUOYANT which can kill eval().
# 2005/09/13
# Added default values for *_GetVar() functions, to be returned when
# variables have not been created.  Also added setToDefaults flag to
# create (and set to the given default value) those previously uncreated
# variables.
# 2005/10/01
# Shaved some time off of the basic Get_/Set_ calls by using functional
# tools.
# 2006/09/28
# Some minor changes for use in FFX 3.2.  Probably more to come
# 2006/10/10
# Major changes with use of _missionvar.pyd compiled module for speedup
# of base functions and attempt at caching.  My testing shows write-through
# aching has a big benefit for mission variables but is worse for object
# variables.  Therefore, I am assigning the write-through ("_WT") functions
# to the mission work and calling the _missionvar module stuff for objects.
# Meanwhile, there is no reason to write to attribute space at all for
# Rumble Room sessions.  So, I am assigning the simulated (via the caching
# dictionaries) routines for use in RR play.
# 2006/10/14
# fixed problem where cached GetVar calls wouldn't accept None as a default
# value
# 2006/10/14
# Figured out how to determine which saved game we just loaded from
# and read attribute data from that.  This allows us to do per object
# instead of (much-more-costly) per variable write-through caching for
# object variables in a saved-game safe manner.  The *_GetVar()
# functions now assume that the cache is "clean"; that is, if a read
# doesn't find a variable in the cache, we don't check the attributes.
# 2006/12/24
# Changed way functions are defined for skirmish/campaign play to 
# check once here at module load and then check again if sk.py is loaded.
# Note that this requires a small change to sk.py as well as being 
# carefull in ffx.py not to load skirmish code unless someone calls the
# skirmish init.
# 2007/10/19 v1.19
# Removed import of datfiles from code. Not needed now.
# 2008/04/19 v1.20
# Fixed name bug for object retirement that Epimethee noticed.

############################

print "importing missionobjvar.py v1.20"

#########################################################################
# imports
#########################################################################

if ( __name__ != '__main__' ) and ( __name__ != '__cshelper__' ):
    import js
    import ff
    import types
    ##import datfiles
    import mlogreader
    import readsavedattribs
    ##import _missionvar
    from js import *
    from ff import *

#########################################################################
# (set up the functions to use based on mission type at bottom of file)
#########################################################################

#########################################################################
# Just a utility function
#########################################################################

def Mission_AttrExists(attrName):
    try:
        dummy = Mission_GetAttr(attrName)
        return 1
    except:
        return 0

#########################################################################
# Caching data (see CleanVarCache() later in this file)
#########################################################################

__MISSIONVARS = {}
__OBJECTVARS = {}

#########################################################################
# NoVal
# We define something to represent no specified variable passed to a
# function. This is so that we can distinguish from someone not
# passing a default return value to the *_GetVar functions and someone
# passing None as the default value.
#########################################################################

class NoValClass:
    pass

NoVal = NoValClass()

#########################################################################
# The following version uses write-through caching 
#########################################################################

# Set a mission long var
def Mission_SetVar_WT(varName,val):
    _missionvar._Mission_SetVar(varName,val)
    __MISSIONVARS[varName] = val
                       
# Retrieve a mission long var
def Mission_GetVar_WT(varName, defaultVal=NoVal, setToDefault=0):
    if __MISSIONVARS.has_key(varName):
        return __MISSIONVARS[varName]
    if defaultVal != NoVal:
        if setToDefault:
            _missionvar._Mission_SetVar(varName,defaultVal)
            __MISSIONVARS[varName] = defaultVal
        return defaultVal
    print "Mission_GetVar called for nonexistant var %s." % repr(varName)
    raise RuntimeError

# Does the mission long var exist?
def Mission_VarExists_WT(varName):
    if __MISSIONVARS.has_key(varName):
        return 1
    return 0

#These object functions set an OBJECT_DESTROY sink for each variable
# when it is first created.

# Retrieve an object long var
def Object_SetVar_WT(object,varName,val):
    if Object_Exists(object):
        _missionvar._Object_SetVar(object,varName,val,RemoveOnDestroy=0)
        if not __OBJECTVARS.has_key(object):
            __OBJECTVARS[object] = { varName : val }
            mlogreader.regObjectDestroy(object, 'missionobjvar.Object_onRemoveObjectFromList_WT')
        else:
            __OBJECTVARS[object][varName] = val
    else:
        __Object_RemoveObjectFromVarList_WT(object)  # just in case this was a destroyed object
        print 'Object_SetVar: object <%s> does not exist; %s cannot be set.' % (repr(object),repr(varName))
        raise RuntimeError

# Set an object long var
def Object_GetVar_WT(object,varName, defaultVal=NoVal, setToDefault=0):
    if Object_Exists(object):
        if __OBJECTVARS.has_key(object) and __OBJECTVARS[object].has_key(varName):
            return __OBJECTVARS[object][varName]
        if defaultVal != NoVal:
            if setToDefault:
                _missionvar._Object_SetVar(obect,varName,defaultVal,RemoveOnDestroy=0)
                if not __OBJECTVARS.has_key(object):
                    __OBJECTVARS[object] = { varName : defaultVal }
                    mlogreader.regObjectDestroy(object, 'missionobjvar.Object_onRemoveObjectFromList_WT')
                else:
                    __OBJECTVARS[object][varName] = defaultVal
            return defaultVal
        print "Object_GetVar called on object %s for nonexistant var %s." % (repr(object),repr(varName))
        raise RuntimeError
    print "Object_GetVar called on nonexistant object %s." % (repr(object),)
    raise RuntimeError

# Does the object have a particular long var?
def Object_VarExists_WT(object,varName):
    if __OBJECTVARS.has_key(object) and __OBJECTVARS[object].has_key(varName):
        return 1
    return 0

#########################################################################
# Object variable retirement and cache cleanup for write-through
# schemes
#########################################################################

# Clean up data when an object has been destroyed, do every variable for the object all at once
def Object_onRemoveObjectFromList_WT(event):
    object = event.object
    __Object_RemoveObjectFromList_WT(object)

def __Object_RemoveObjectFromList_WT(object):
    if __OBJECTVARS.has_key(object):
        varList = __OBJECTVARS[object].keys()
        _missionvar._Object_UnsetList(object,varList)
        del __OBJECTVARS[object]

# Clean up data when an object has been destroyed, do each variable as a separate sink
def Object_onRemoveObjectVarFromList_WT(event):
    object = event.object
    varName = event.string
    __Object_RemoveObjectVarFromList_WT(object,varName)

def __Object_RemoveObjectVarFromList_WT(object,varName):
    if __OBJECTVARS.has_key(object):
        if __OBJECTVARS[object].has_key(varName):
            del __OBJECTVARS[object][varName]
    # We put this here so that only one sink has to be registered (instead of one for the stored var and one for the cached one)
    _missionvar._Object_UnsetVar(object,varName)

# To be called if we discover an object is destroyed.  If we are
# registering the object vars one at a time, this is just backup; the
# object should have had its vars cleaned up by the regObjectDestroy()
# sinks automatically.
# If we are keeping a list of vars for each object, use a different
# function and pass it that list, because this one spends time
# generating the list from the callback mission var.
def __Object_RemoveObjectFromVarList_WT(object):
    if __OBJECTVARS.has_key(object):
        del __OBJECTVARS[object]
    # modified from mlogreader.cancelObjectDestroyCallbackSink() so that we
    # can use that to send a list to _missionvar._Object_UnsetList().
    list = []
    callbacks = missionobjvar.Mission_GetVar('OBJECT_DESTROY_EVENT_CB', defaultVal = [])
    if not callbacks:   # don't make a list if there were no callbacks
        return
    keptcallbacks = []
    for cb in callbacks:
        if ( cb[0] == object ):
            list.append(cb[5])
        else:
            keptcallbacks.append(cb)
    missionobjvar.Mission_SetVar('OBJECT_DESTROY_EVENT_CB', keptcallbacks)
    _missionvar._Object_UnsetList(object,list)
    
#########################################################################
# The following version uses simulated attribute calls.  That is, it's
# ALL caching.  It is for use in the Rumble Room where saved-game
# safety is not an issue.
#########################################################################

# Set a mission long var                       
def Mission_SetVar_S(varName,val):
    __MISSIONVARS[varName] = val
                       
# Retrieve a mission long var
def Mission_GetVar_S(varName, defaultVal=NoVal, setToDefault=0):
    if __MISSIONVARS.has_key(varName):
        return __MISSIONVARS[varName]
    if defaultVal != NoVal:
        if setToDefault:
            __MISSIONVARS[varName] = defaultVal
        return defaultVal
    print "Mission_GetVar called for nonexistant var %s." % repr(varName)
    raise RuntimeError

# Does the mission long var exist?
def Mission_VarExists_S(varName):
    return __MISSIONVARS.has_key(varName)

#These object functions set only one OBJECT_DESTROY sink for each object
# when its first variable is created.

# Retrieve an object long var
def Object_SetVar_S(object,varName,val):
    if Object_Exists(object):
        if not __OBJECTVARS.has_key(object):
            __OBJECTVARS[object] = { varName : val }
            mlogreader.regObjectDestroy(object, 'missionobjvar.__Object_RemoveObjectFromVarList_S')
        else:
            __OBJECTVARS[object][varName] = val
    else:
        __Object_RemoveObjectFromVarList_S(object)  # just in case this was a destroyed object
        print 'Object_SetVar: object <%s> does not exist; %s cannot be set.' % (repr(object),repr(varName))
        raise RuntimeError

# Set an object long var
def Object_GetVar_S(object,varName, defaultVal=NoVal, setToDefault=0):
    if Object_Exists(object):
        if __OBJECTVARS.has_key(object) and __OBJECTVARS[object].has_key(varName):
            return __OBJECTVARS[object][varName]
        if defaultVal != NoVal:
            if setToDefault:
                if not __OBJECTVARS.has_key(object):
                    __OBJECTVARS[object] = { varName : defaultVal}
                    mlogreader.regObjectDestroy(object, 'missionobjvar.__Object_RemoveObjectFromVarList_S')
                else:
                    __OBJECTVARS[object][varName] = defaultVal
            return defaultVal
        print "Object_GetVar called on object %s for nonexistant var %s." % (repr(object),repr(varName))
        raise RuntimeError
    print "Object_GetVar called on nonexistant object %s." % (repr(object),)
    raise RuntimeError

# Does the object have a particular long var?
def Object_VarExists_S(object,varName):
    return __OBJECTVARS.has_key(object) and __OBJECTVARS[object].has_key(varName)

# Clean up data when an object has been destroyed

# To be called if we discover an object is destroyed.  Remove all of its variables from the cache.
# This can be called directly with the object's name or as an event callback
def __Object_RemoveObjectFromVarList_S(object):
    if not isinstance(object,types.StringType):
        object = object.object
    if __OBJECTVARS.has_key(object):
        del __OBJECTVARS[object]

#########################################################################
# Cleaning a "dirty" cache after a saved-game is loaded
#########################################################################

# This function reads the attribute data stored in the given
# saved-game file and uses it to populate the cache after a saved-game
# load.  It is called by mlogreader's MLOG_Read() function, which
# should be among the first to notice that a saved game has been
# loaded.
#
# I am trying to be careful here, in case some long variable data has
# been stored since the saved game was loaded.  So any data that is in
# the cache is assumed to be new.  So, the stuff read from the saved
# game is only added if it's missing from the current cache.
def CleanVarCache(FileName):
    # read and unpack the mission and object variables
    MV, OV = readsavedattribs.ReadLongVarsFromSavedGame(FileName)
    # Update the mission var cache dictionary
    CurrentKeys = __MISSIONVARS.keys()
    for key in filter(lambda k, CKs=CurrentKeys: k not in CKs,MV.keys()):
        __MISSIONVARS[key] = MV[key]
    # Update the object var cache dictionary
    CurrentObjects = __OBJECTVARS.keys()
    for object in OV.keys():
        if object not in CurrentObjects:    # whole new object
            __OBJECTVARS[object] = OV[object]
        else:                               # just new keys
            CurrentKeys = __OBJECTVARS[object].keys()
            for key in filter(lambda k, CKs=CurrentKeys: k not in CKs,OV[object].keys()):
                __OBJECTVARS[object][key] = OV[object][key]

# Just for testing...
# NB: This is a poor test, since it won't warn about empty caches.
def ConfirmCache():
    for var in __MISSIONVARS.keys():
        if _missionvar._Mission_GetVar(var) != __MISSIONVARS[var]:
            print 'ConfirmCache: PROBLEM!! %s differs from %s' % (_missionvar._Mission_GetVar(var),__MISSIONVARS[var])
    for object in __OBJECTVARS.keys():
        for var in __OBJECTVARS[object].keys():
            if _missionvar._Object_GetVar(object,var) != __OBJECTVARS[object][var]:
                print 'ConfirmCache: PROBLEM!! %s differs from %s' % (_missionvar._Mission_GetVar(var),__MISSIONVARS[var])

#########################################################################
# original low-level calls, replaced by _missionvar module
#########################################################################

# Lower-level stuff: does not use hash tables; less memory efficient for
# longer variable names.  NOTE: These functions are used by the ones above
# but are not compatible.  Using both in the same app is asking for trouble.
import thread

def __Mission_VarExists(varName):
    try:
        len = Mission_GetAttr(varName+'_len')
        return 1
    except:
        return 0

def __Mission_SetVar(varName,val):
    valstr = repr(val)
    strlen = len(valstr)
    numlist = PackToNumList(valstr)
    mutex = thread.allocate_lock()  # don't interrupt this next
    mutex.acquire()
    Mission_SetAttr(varName+'_len',strlen)
    attrnames = map(lambda i, varName=varName: varName + '_%x' % i, range(len(numlist)))
    map(Mission_SetAttr,attrnames,numlist)
    mutex.release()
                       
def __Mission_GetVar(varName):
    if __Mission_VarExists(varName):
        strlen = int(Mission_GetAttr(varName+'_len'))
        attrnames = map(lambda i, varName=varName: varName + '_%x' % i, range((strlen+FLOATLEN-1) / FLOATLEN))
        numlist = map(Mission_GetAttr,attrnames)
        valstr = UnpackToString(numlist)[:strlen]
##        print 'valstr=%s' % repr(valstr)
        val = eval(valstr)
        return val
    else:
        return None

def __Object_SetVar(object,varName,val):
    valstr = repr(val)
    strlen = len(valstr)
    numlist = PackToNumList(valstr)
    listlen = len(numlist)
    mutex = thread.allocate_lock()  # don't interrupt this next
    mutex.acquire()
    Object_SetAttr(object,varName+'_len',strlen)
    attrnames = map(lambda i, varName=varName: varName + '_%x' % i, range(listlen))
    map(Object_SetAttr,[object]*listlen,attrnames,numlist)
    mutex.release()
                       
def __Object_GetVar(object,varName):
    strlen = int(Object_GetAttr(object,varName+'_len'))
    if strlen == 0:
        print '__Object_GetVar(%s,%s) failed to get strlen' % (object,varName)
        return None
    listlen = (strlen+FLOATLEN-1) / FLOATLEN
    attrnames = map(lambda i, varName=varName: varName + '_%x' % i, range(listlen))
    numlist = map(Object_GetAttr,[object]*listlen,attrnames)
    valstr = UnpackToString(numlist)[:strlen]
    val = eval(valstr)
    return val


#########################################################################
# The routines to pack a string into a list of numbers
#########################################################################

import struct
FLOATLEN=struct.calcsize('f')  # how many bytes will we store in a float?

try:
    ## best version
    import chunkstring    # My compiled C module, if available
    def PackToNumList(s):
        return chunkstring.string2numtuple(s)

    def UnpackToString(numlist):
        return chunkstring.numtuple2string(tuple(numlist))

except:
    ## decent version
    # Turn a string into a sequence of long integers
    def PackToNumList(s):
        fl = FLOATLEN
        slen = len(s)
        r = slen - (slen/fl)*fl
        if r != 0:  s = s + '\x00'*(fl-r)
        num = []
        for i in range(0,slen,fl):
            num.append(struct.unpack('f',s[i:i+fl])[0])
        return num

    def UnpackToString(numlist):
        s = ''
        for num in numlist:
            s = s + struct.pack('f',num)
        return s

#########################################################################
# Clean up poorly exported longs
#########################################################################

# Make sure that any long ints exported from the game engine modules
# weren't exported with normal int object types, because that will
# mess up the eval() statement in the *_GetVar() functions.  I think
# SCSTATE_BUOYANT is the only culprit, but this should catch any
# others.
# Can be called without args and also with a list of pairs of variable
# names and values to look at.  The item names in that list should
# have the module names prepended, otherwise they will be evaluated in
# this module.  It is okay to pass the values 'broken'.
# Example:
#  FixLongs([('cshelper.SCSTATE_BUOYANT',cshelper.SCSTATE_BUOYANT),('ffx.SCSTATE_BUOYANT',ffx.SCSTATE_BUOYANT)])
def FixLongs(varlist=None):
    import types
    if not varlist:
        ints = filter(lambda x,globs=globals(),IT=types.IntType: isinstance(globs[x],IT),globals().keys())
        for i in ints:
            try:
                eval(repr(globals()[i]))
            except OverflowError:
                print 'missionobjvar(FixLongs): Fixing overflow issue with <%s>' % i
                fixstr = 'js.%s = %dL' % (i,globals()[i])
                exec fixstr in globals()
            except:
                print 'missionobjvar(FixLongs): other problem with %s' % i
    else:
        import string
        for i in varlist:
            try:
                ##print 'missionobjvar(FixLongs): about to eval <%s>' % i[0]
                modname = i[0][:string.rfind(i[0],'.')]
                ##print 'missionobjvar(FixLongs): running <%s>' % ('import ' + modname)
                exec 'import ' + modname
                print 'missionobjvar(FixLongs): %s = %d' % (i[0],eval(repr(i[1])))
            except OverflowError:
                print 'missionobjvar(FixLongs): Fixing overflow issue with <%s>' % i[0]
                fixstr = '%s = %dL' % (i[0],i[1])
                d = eval(modname+'.__dict__')
                ##print 'missionobjvar(FixLongs): fixstr = %s' % (repr(fixstr))
                exec fixstr in d
            except:
                print 'missionobjvar(FixLongs): other problem with %s -  %s  %s' % (i, sys.exc_type, sys.exc_value)

# Now run this on module load and somewhere outside (in ffx.py) to
# catch modules that loaded before it.
FixLongs()

#########################################################################
# set up the functions to use based on mission type
#########################################################################

# Is there a good way to do this during an early import like this?
# This will fail for RR missions that have the same names as campaign
# missions.  But that is a safe failure (meaning that we use the
# conservative variable code that is saved-game safe).

# stumpy 2006-12-24: *** Another option here is to write a function
# that defines the functions (which would need to be declared as
# global within that defining function). We do the isCampaign() check
# here when this module loads (which sometimes marks the RR as a
# campaign) and then call the defining function from sk.py as well,
# redefining this module's functions.
#
# Will that cause crashes? It should't if none of the missionobjvar
# long var functions have been called yet. We could check the lengths
# of __MISSIONVARS and __OBJECTVARS to see if they are still zero. I
# think there is a good possibility that going from the write-through
# scheme to the simulated scheme (which is the only case likely to
# occur) wouldn't cause trouble anyway...

def isCampaign():
    try:
        test = Mission_GetAttr('Campaign_Mission')
        return test
    except RuntimeError:
        return 1

def SetFunctionsByMissionType(isCampaignMission=1):
    global _missionvar
    global Mission_SetVar, Mission_GetVar, Mission_VarExists, Object_SetVar, Object_GetVar, Object_VarExists
    # do a little test here to see if we've already started writing lang vars
    if __MISSIONVARS or __OBJECTVARS:
        print 'missionobjvar.SetFunctionsByMissionType: isCampaignMission=%d, long variable cache has already been written to.' % isCampaignMission
        print '__MISSIONVARS = %s' % __MISSIONVARS
        print '__OBJECTVARS = %s' % __OBJECTVARS
    if isCampaignMission:
        print "missionobjvar defining functions for Campaign play."
        import _missionvar
        Mission_SetVar = Mission_SetVar_WT
        Mission_GetVar = Mission_GetVar_WT
        Mission_VarExists = Mission_VarExists_WT
        Object_SetVar = Object_SetVar_WT
        Object_GetVar = Object_GetVar_WT
        Object_VarExists = Object_VarExists_WT
    else:
        print "missionobjvar defining functions for Rumble Room play."
        Mission_SetVar = Mission_SetVar_S
        Mission_GetVar = Mission_GetVar_S
        Mission_VarExists = Mission_VarExists_S
        Object_SetVar = Object_SetVar_S
        Object_GetVar = Object_GetVar_S
        Object_VarExists = Object_VarExists_S

SetFunctionsByMissionType(isCampaign())