#! /usr/bin/env python
#
# readsavedattribs.py
#
# More parsing through DAT files, this time the saved game file to
# grab mission and object attributes. This is useful for diagnostic
# purposes and also in-game for refereshing the mission and object
# long variable caches, allowing saved-game safe write-through
# caching.
#
# The diagnostic tools (near the end of the module) allow us to see
# the mission variables, the object variables, the FFX object
# variables, and the remaining mission and object pure attributes.
# This latter are the object attributes minus those used to store
# object long variables and mission attributes minus those used to
# store object long variables and those used to store FFX object
# attributes.
#
#
# (c) stumpy 2006, 2007


import struct
import string

#########################################################################
# The routines to pack a string into a list of numbers
#########################################################################

FLOATLEN=struct.calcsize('f')  # how many bytes will we store in a float?
try:
    ## best version
    import chunkstring    # My compiled C module, if available
    def PackToNumList(s):
        return chunkstring.string2numtuple(s)

    def UnpackToString(numlist):
        return chunkstring.numtuple2string(tuple(numlist))

except:
    ## decent version
    # Turn a string into a sequence of long integers
    def PackToNumList(s):
        fl = FLOATLEN
        slen = len(s)
        r = slen - (slen/fl)*fl
        if r != 0:  s = s + '\x00'*(fl-r)
        num = []
        for i in range(0,slen,fl):
            num.append(struct.unpack('f',s[i:i+fl])[0])
        return num

    def UnpackToString(numlist):
        s = ''
        for num in numlist:
            s = s + struct.pack('f',num)
        return s

#########################################################################
# Counting in base 36 - fun for the whole family!
#########################################################################

# only good for ints up to 36**5-1
def itoa36(num):
    digits = '0123456789abcdefghijklmnopqrstuvwxyz'
    s = ''
    div = num/1679616
    if div > 0:     # 36**4
        s = s + digits[div]
        num = num % 1679616
    div = num/46656 
    if div > 0:     # 36**3
        s = s + digits[div]
        num = num % 46656
    elif s:
        s = s + '0'
    div = num/1296 
    if div > 0:     # 36**2
        s = s + digits[div]
        num = num % 1296
    elif s:
        s = s + '0'
    div = num/36 
    if div > 0:     # 36**1
        s = s + digits[div]
        num = num % 36
    elif s:
        s = s + '0'
    s = s + digits[num]
    return s

#########################################################################
# getting attributes from a saved game file
#########################################################################

def GrabGlobalAttr(dat,startpos):
    len = ord(dat[startpos])
    name = dat[startpos+4:startpos+4+len]
    val = struct.unpack('f',dat[startpos+len+4:startpos+len+8])[0]
    return (startpos+len+28,{name:val})

# return a dictionary of all the attributes as set by GlobalRegisterAttr()
def GetGlobalAttributes(dat,verbose=0):
    incr = 28
    # The penultimate 'CSymTab' occurrence
    startpos = string.rfind(dat,'CSymTab')
    numatts = struct.unpack('H',dat[startpos+15:startpos+17])[0]
    startpos = startpos + 19
    attnum = 0
    attribs = {}
    while (attnum < numatts):
        len = ord(dat[startpos])
        name = dat[startpos+4:startpos+4+len]
        if '\0' in name:
            print 'possible bad attribute at %d' % startpos
        val = struct.unpack('f',dat[startpos+len+4:startpos+len+8])[0]
        if verbose:
            print '%10d\t%e\t%s' % (startpos,val,name)
        attribs[name] = val
        attnum = attnum + 1
        startpos = startpos + len + incr
    return attribs

# return a dictionary of all the attributes as set by Mission_SetAttr()
def GetMissionAttributes(dat,verbose=0):
    incr = 28
    # The penultimate 'CSymTab' occurrence
    startpos = string.rfind(dat,'CSymTab')
    startpos = string.rfind(dat[:startpos],'CSymTab')
    numatts = struct.unpack('H',dat[startpos+15:startpos+17])[0]
    startpos = startpos + 19
    attnum = 0
    attribs = {}
    while (attnum < numatts):
        len = ord(dat[startpos])
        name = dat[startpos+4:startpos+4+len]
        if '\0' in name:
            print 'GetMissionAttributes: possible bad attribute at %d' % startpos
        val = struct.unpack('f',dat[startpos+len+4:startpos+len+8])[0]
        if verbose:
            print 'GetMissionAttributes: %10d\t%e\t%s' % (startpos,val,name)
        attribs[name] = val
        attnum = attnum + 1
        startpos = startpos + len + incr
    return attribs

# return a dictionary of all the attributes as set by Mission_SetAttr()
def GetObjectAttributes(dat, object, verbose=0):
    incr = 28
    # The penultimate 'CSymTab' occurrence
    startpos = string.rfind(dat,'_GameObjData'+object+'\0')
    startpos = string.find(dat,'CSymTab',startpos)
    numatts = struct.unpack('H',dat[startpos+15:startpos+17])[0]
    startpos = startpos + 19
    attnum = 0
    attribs = {}
    while (attnum < numatts):
        len = ord(dat[startpos])
        name = dat[startpos+4:startpos+4+len]
        if '\0' in name:
            print 'possible bad attribute at %d' % startpos
        val = struct.unpack('f',dat[startpos+len+4:startpos+len+8])[0]
        if verbose:
            print 'GetObjectAttributes: %10d\t%e\t%s' % (startpos,val,name)
        attribs[name] = val
        attnum = attnum + 1
        startpos = startpos + len + incr
    if verbose:
        print 'GetObjectAttributes: got %d attributes for %s' % (attnum,object)
    return attribs

# separate out the normal mission attributes from the long variable attributes
def GetMissionVars(rawattribs):
    missionvars = filter(lambda att: att[-4:]=='*len',rawattribs.keys())   # the length headers
    mvars = {}
    objectvarnames = {}
    for mvar in missionvars:
        basename = mvar[:-4]
        strlen = int(rawattribs[mvar])
        if '*' in basename: # this is the len entry for an object varriable
            object, varname = string.split(basename,'*')
            ##print 'mvar=%s, basename=%s, strlen=%d, object var %s found for %s' % (repr(mvar),repr(basename),strlen,repr(varname),repr(object))
            if objectvarnames.has_key(object):
                objectvarnames[object][varname] = strlen
            else:
                objectvarnames[object] = { varname : strlen }
            del rawattribs[mvar]
            continue
        if strlen == 0:  # plain numerical value
            attname = basename+'*0'
            mvars[basename] = rawattribs[attname]
            del rawattribs[attname]
            continue
        elif strlen < 0:    # this is a "retired" mission variable
            continue
        numatts = (strlen+3)/4
        ##print 'mvar=%s, basename=%s, strlen=%d, numatts=%d' % (repr(mvar),repr(basename),strlen,numatts)
        attlist = []
        for att in range(numatts):
            attname = basename+'*'+itoa36(att)
            ##print '  getting %s = %e (%s)' % (attname,rawattribs[attname],struct.pack('f',rawattribs[attname]))
            attlist.append(rawattribs[attname])
            del rawattribs[attname]
        mvars[basename] = eval(UnpackToString(attlist)[:strlen])
        del rawattribs[mvar]
    return mvars, objectvarnames

# Get the object long variables for all vars and return it as a dictionary
def GetAllObjectVars(dat,objectvarnames):
    ##print 'GetAllObjectVars: objectvarnames=%s' % (objectvarnames,)
    objectvars = {}
    for object in objectvarnames.keys():
        objectvars[object] = {}
        rawattribs = GetObjectAttributes(dat,object)
        ##print 'GetAllObjectVars: (%s) rawattribs=%s' % (object, rawattribs)
        for ovar in objectvarnames[object].keys():
            strlen = objectvarnames[object][ovar]
            if strlen == 0:     # plain numerical value
                ##print 'GetAllObjectVars: (%s) ovar=%s, strlen=%d NUMERICAL' % (object, repr(ovar),strlen)
                attname = ovar+'*0'
                objectvars[object][ovar] = rawattribs[attname]
                del rawattribs[attname]
                continue
            elif strlen < 0:    # this is a "retired" object variable
                ##print 'GetAllObjectVars: (%s) ovar=%s, strlen=%d RETIRED' % (object, repr(ovar),strlen)
                continue
            numatts = (strlen+3)/4
            ##print 'GetAllObjectVars: (%s) ovar=%s, strlen=%d, numatts=%d' % (object, repr(ovar),strlen,numatts)
            attlist = []
            for att in range(numatts):
                attname = ovar+'*'+itoa36(att)
                ##print '  GetAllObjectVars: getting %s = %e (%s)' % (attname,rawattribs[attname],repr(struct.pack('f',rawattribs[attname])))
                attlist.append(rawattribs[attname])
                del rawattribs[attname]
            objectvars[object][ovar] = eval(UnpackToString(attlist)[:strlen])
    return objectvars

# Return a dictionary of all mission vars and a dictionary of all
# object vars from the given saved game file.  This is assuming
# variables stored the way they are in _missionvars module, as used in
# missionobjvars version 1.15 and later.
#
# This routine does no caching, in case a later read is requested from
# a file saved later under the same name (e.g. a second quicksave).
def ReadLongVarsFromSavedGame(FileName):
    f = open(FileName,'rb')
    alldat = f.read()
    f.close()
    MissionAttributes = GetMissionAttributes(alldat)
    MissionVars, ObjectVarNames = GetMissionVars(MissionAttributes)
    ObjectVars = GetAllObjectVars(alldat,ObjectVarNames)
    return MissionVars, ObjectVars

# (Same as above function, but assumes that the binary data string
# from the saved-game file is passed.  Somewhat more efficient if you
# happen to have the data already read.)
#
# Return a dictionary of all mission vars and a dictionary of all
# object vars from the given binary string from a saved game file.
# This is assuming variables stored the way they are in _missionvars
# module, as used in missionobjvars version 1.15 and later.
#
# This routine does no caching, in case a later read is requested from
# a file saved later under the same name (e.g. a second quicksave).
def ReadLongVarsFromSavedGameData(alldat):
    MissionAttributes = GetMissionAttributes(alldat)
    MissionVars, ObjectVarNames = GetMissionVars(MissionAttributes)
    ObjectVars = GetAllObjectVars(alldat,ObjectVarNames)
    return MissionVars, ObjectVars

############################
# diagnostic tools
############################

# Separate out the normal mission attributes from the long variable
# attributes.
#
# If RemoveFFXAttribs is set, also remove from the "normal mission
# attributes" those being used to store FFX "object" attributes. This
# last bit is something of a guessing game, since an attribute like
# 'hero_4_static' might be an FFX attribute 'static' for the 'hero_4'
# object or it really might just be a mission attribute with that
# name. And, it will not remove the FFX attribs for objects that no
# longer exist, which can be numerous.
#
# Note that none of this is written for efficiency and removal of the
# FFX attributes is particularly slow.
def GetAllMissionVarsAndAttribs(FileName,RemoveFFXAttribs=0):
    f = open(FileName,'rb')
    alldat = f.read()
    f.close()
    rawattribs = GetMissionAttributes(alldat)
    if RemoveFFXAttribs:    # now remove the FFX object attribs
        # first get the names of all stored objects
        pos=0
        allobjs = []
        while pos != -1:
            pos = string.find(alldat,'_GameObjData',pos)
            if pos != -1:
                pos = pos+12
                object = alldat[pos:string.find(alldat,'\0',pos,pos+256)]
                allobjs.append(object)
                ##print '%s found at %d' % (object,pos)
                pos = pos + 1
        # find any likely FFX object attributes stored in mission attributes
        for obj in allobjs:
            ##print 'looking at %s' % obj
            attstart = '%s_' % obj
            ffxatts = filter(lambda att: string.find(att,attstart)==0,rawattribs.keys())
            for ffxatt in ffxatts:
                ## print '  removing ffxattrib %s' % ffxatt[len(attstart):]
                del rawattribs[ffxatt]
    missionvars = filter(lambda att: att[-4:]=='*len',rawattribs.keys())   # the length headers
    mvars = {}
    objectvarnames = {}
    for mvar in missionvars:
        basename = mvar[:-4]
        strlen = int(rawattribs[mvar])
        if '*' in basename: # this is the len entry for an object varriable
            object, varname = string.split(basename,'*')
            ##print 'mvar=%s, basename=%s, strlen=%d, object var %s found for %s' % (repr(mvar),repr(basename),strlen,repr(varname),repr(object))
            if objectvarnames.has_key(object):
                objectvarnames[object][varname] = strlen
            else:
                objectvarnames[object] = { varname : strlen }
            del rawattribs[mvar]
            continue
        if strlen == 0:  # plain numerical value
            attname = basename+'*0'
            mvars[basename] = rawattribs[attname]
            del rawattribs[attname]
            continue
        elif strlen < 0:    # this is a "retired" mission variable
            continue
        numatts = (strlen+3)/4
        ##print 'mvar=%s, basename=%s, strlen=%d, numatts=%d' % (repr(mvar),repr(basename),strlen,numatts)
        attlist = []
        for att in range(numatts):
            attname = basename+'*'+itoa36(att)
            ##print '  getting %s = %e (%s)' % (attname,rawattribs[attname],struct.pack('f',rawattribs[attname]))
            attlist.append(rawattribs[attname])
            del rawattribs[attname]
        mvars[basename] = eval(UnpackToString(attlist)[:strlen])
        del rawattribs[mvar]
    return mvars, rawattribs

# Get the object long variables for a particular object and return it
# as a dictionary, along with the remaining object attribs and FFX
# "object" attributes.
def GetAllObjectVarsAndAttribs(FileName,ObjectName):
    f = open(FileName,'rb')
    dat = f.read()
    f.close()
    MissionAttributes = GetMissionAttributes(dat)
    # find FFX attributes
    ffxobjectattribs = {}
    attstart = '%s_' % ObjectName
    ffxatts = filter(lambda att: string.find(att,attstart)==0,MissionAttributes.keys())
    for ffxatt in ffxatts:
        ##print '  getting ffxattrib %s' % ffxatt[len(attstart):]
        ffxobjectattribs[ffxatt[len(attstart):]] = MissionAttributes[ffxatt]
        del MissionAttributes[ffxatt]
    # Now find the mission variables
    MissionVars, objectvarnames = GetMissionVars(MissionAttributes)
    ##print 'GetAllObjectVars: objectvarnames=%s' % (objectvarnames,)
    objectvars = {}
    rawattribs = GetObjectAttributes(dat,ObjectName)
    ##print 'GetAllObjectVars: (%s) rawattribs=%s' % (ObjectName, rawattribs)
    if objectvarnames.has_key(ObjectName):
        for ovar in objectvarnames[ObjectName].keys():
            strlen = objectvarnames[ObjectName][ovar]
            if strlen == 0:     # plain numerical value
                ##print 'GetAllObjectVars: (%s) ovar=%s, strlen=%d NUMERICAL' % (ObjectName, repr(ovar),strlen)
                attname = ovar+'*0'
                objectvars[ovar] = rawattribs[attname]
                del rawattribs[attname]
                continue
            elif strlen < 0:    # this is a "retired" ObjectName variable
                ##print 'GetAllObjectVars: (%s) ovar=%s, strlen=%d RETIRED' % (ObjectName, repr(ovar),strlen)
                continue
            numatts = (strlen+3)/4
            ##print 'GetAllObjectVars: (%s) ovar=%s, strlen=%d, numatts=%d' % (ObjectName, repr(ovar),strlen,numatts)
            attlist = []
            for att in range(numatts):
                attname = ovar+'*'+itoa36(att)
                ##print '  GetAllObjectVars: getting %s = %e (%s)' % (attname,rawattribs[attname],repr(struct.pack('f',rawattribs[attname])))
                attlist.append(rawattribs[attname])
                del rawattribs[attname]
            objectvars[ovar] = eval(UnpackToString(attlist)[:strlen])
    return objectvars, rawattribs, ffxobjectattribs

# Skeleton code for use in clearing FFX object attributes out of mission attribute list
##    # first get the names of all stored objects
##    pos=0
##    allobjs = []
##    while pos != -1:
##        pos = string.find(dat,'_GameObjData',pos)
##        if pos != -1:
##            pos = pos+12
##            object = dat[pos:string.find(dat,'\0',pos,pos+256)]
##            allobjs.append(object)
##            ##print '%s found at %d' % (object,pos)
##            pos = pos + 1
##    # find any likely FFX object attributes stored in mission attributes
##    ffxobjectattribs = {}
##    for obj in allobjs:
##        print 'looking at %s' % obj
##        attstart = '%s_' % obj
##        ffxatts = filter(lambda att: string.find(att,attstart)==0,MissionAttributes.keys())
##        if ffxatts:
##            ffxobjectattribs[obj] = {}
##        for ffxatt in ffxatts:
##            print '  getting ffxattrib %s' % ffxatt[len(attstart):]
##            ffxobjectattribs[obj][ffxatt[len(attstart):]] = MissionAttributes[ffxatt]
##            del MissionAttributes[ffxatt]
